@@ -2,6 +2,7 @@
 
 from __future__ import annotations
 
+import os
 import subprocess
 import sys
 
@@ -20,11 +21,16 @@ class TestMainModule:
     def test_main_module_executes(self):
         """Test that running the module as main executes correctly."""
         # Use subprocess to run the module as __main__ and check it doesn't crash
-        # We expect it to show help/error since we're not providing arguments
+        # Use --version flag for a quick, deterministic test that doesn't require user input
+        env = {**os.environ, "HUD_SKIP_VERSION_CHECK": "1"}
         result = subprocess.run(
-            [sys.executable, "-m", "hud.cli"], capture_output=True, text=True, timeout=10
+            [sys.executable, "-m", "hud.cli", "--version"],
+            capture_output=True,
+            text=True,
+            timeout=30,
+            env=env,
         )
 
-        # Should exit with an error code but not crash
-        # (The actual main function will show help or error for missing args)
-        assert result.returncode != 0  # CLI should exit with error for no args
+        # Should exit successfully with version info
+        assert result.returncode == 0
+        assert "version" in result.stdout.lower() or "hud" in result.stdout.lower()
